@@ -1,20 +1,20 @@
 from warnings import warn
 from s3fs import S3FileSystem
 from time import sleep
-# from botocore.exceptions import NoCredentialsError
+#  from botocore.exceptions import NoCredentialsError
 from pandas import read_csv
 from pandas import DataFrame as PandasDF
 from pickle import dump as pickle_dump
 from pickle import load as pickle_load
 from psycopg2 import connect as psycopg2_connect
 from csv import QUOTE_NONNUMERIC
-from pyspark.sql.dataframe import DataFrame as SparkDF
-from pyspark.sql import SparkSession
+#  from pyspark.sql.dataframe import DataFrame as SparkDF
+#  from pyspark.sql import SparkSession
 from .S3File import S3Files
 
 
 class S3:
-	def __init__(self, key=None, secret=None, iam_role=None, root='s3://', spark=None, **kwargs):
+	def __init__(self, key=None, secret=None, iam_role=None, root='s3://', spark=None):
 		"""
 		starts an S3 connection
 		:type key: str or NoneType
@@ -23,8 +23,6 @@ class S3:
 		:type root: str or NoneType
 		:type spark: pyspark.sql.session.SparkSession or NoneType
 		"""
-		if spark is None:
-			spark = SparkSession.builder.getOrCreate()
 
 		self._key = key
 		self._secret = secret
@@ -112,13 +110,12 @@ class S3:
 				if try_number > 1:
 					print(f'S3 credential issue solved after {try_number} attempts.')
 				return result
-			except: # NoCredentialsError as exception:
+			except: #  NoCredentialsError as exception:
 				print(f'S3 credential issue. Attempt {try_number} failed!')
 				sleep(2 ** (try_number - 1) * 0.2)
 
 		raise exception
 
-
 	def write_bytes(self, path, bytes):
 		path = self._get_absolute_path(path)
 		with self.file_system.open(path=path, mode='wb') as f:
@@ -131,15 +128,21 @@ class S3:
 	def write_csv(self, data, path, index=False, encoding='utf-8', **kwargs):
 		"""
 
-		:type data: SparkDF
+		:type data: pyspark.sql.DataFrame or Pandas.DataFrame
 		:type path: str
 		:type index: bool
 		:type encoding: str
 		:rtype:
 		"""
+
 		path = self._get_absolute_path(path)
-		_bytes = data.to_csv(path_or_buf=None, quoting=QUOTE_NONNUMERIC, index=index, **kwargs).encode(encoding)
-		return self.write_bytes(path=path, bytes=_bytes)
+		if isinstance(data, PandasDF):
+			_bytes = data.to_csv(path_or_buf=None, quoting=QUOTE_NONNUMERIC, index=index, **kwargs).encode(encoding)
+			return self.write_bytes(path=path, bytes=_bytes)
+		else:
+			from pyspark.sql import DataFrame as SparkDF
+			if isinstance(data, SparkDF):
+				return data.write.csv(path=path, encoding=encoding, **kwargs)
 
 	def read_bytes(self, path):
 		with self.file_system.open(path=self._root+path, mode='rb') as f:
@@ -219,7 +222,7 @@ class S3:
 	def write_parquet(self, data, path, mode='overwrite'):
 		"""
 		saves a Spark DataFrame to a path on S3 and returns the list of parquet files
-		:type data: SparkDF
+		:type data: pyspark.sql.DataFrame
 		:type path: str
 		:type mode: str
 		:rtype: list[str]
@@ -238,7 +241,8 @@ class S3:
 		:rtype: pyspark.sql.session.SparkSession
 		"""
 		if self._spark is None:
-			return RuntimeError('S3 does not have access to a SparkSession')
+			from pyspark.sql.session import SparkSession
+			self._spark = SparkSession.builder.getOrCreate()
 		return self._spark
 
 	def set_spark(self, spark):
@@ -284,6 +288,9 @@ class S3:
 	def is_dir(self, path):
 		return self.file_system.isdir(path)
 
+	def rename(self, path1, path2):
+		self.file_system.rename(path1=path1, path2=path2)
+
 	@property
 	def json(self):
 		return self.file_system.to_json()
@@ -336,41 +343,23 @@ class S3:
 
 			return result
 
-	list = ls
-	move = mv
-	dir = ls
-	copy = cp
-	delete = rm
-	load_parquet = read_parquet
-	save_parquet = write_parquet
-	load_pickle = read_pickle
-	save_pickle = write_pickle
-
 	def save(self, obj, path, mode='overwrite'):
 		"""
 		writes an object to path as a parquet if the object is a Spark DataFrame or a pickle otherwise
-		:type obj: object
+		:type obj: object or pyspark.sql.DataFrame
 		:type path: str
 		:type mode: str
 		:rtype: bool
 		"""
-		if isinstance(obj, SparkDF):
-			if path.endswith('.pickle'):
-				raise ValueError(f'Spark DataFrame cannot be saved into a pickle file: "{path}"')
-
-			if not path.endswith('.parquet'):
-				path += '.parquet'
+		if path.endswith('.pickle'):
+			self.write_pickle(obj=obj, path=path, mode=mode)
 
+		elif path.endswith('.parquet'):
 			self.write_parquet(data=obj, path=path, mode=mode)
 
 		else:
-			if path.endswith('.parquet'):
-				raise ValueError(f'{type(path)} cannot be saved into a parquet file: "{path}"')
-
-			if not path.endswith('.pickle'):
-				path += '.pickle'
-
-			self.write_pickle(obj=obj, path=path, mode=mode)
+			ext = path.split('.')[-1]
+			raise ValueError(f'Unsupported file type: {ext}')
 
 		return path
 
@@ -406,6 +395,16 @@ class S3:
 	def __add__(self, other):
 		return self.__truediv__(other)
 
+	list = ls
+	move = mv
+	dir = ls
+	copy = cp
+	delete = rm
+	load_parquet = read_parquet
+	save_parquet = write_parquet
+	load_pickle = read_pickle
+	save_pickle = write_pickle
+
 
 class S3Path:
 	def __init__(self, s3, path):
@@ -468,12 +467,6 @@ class S3Path:
 		"""
 		return self.s3.spark
 
-	def set_spark(self, spark):
-		"""
-		:type spark: pyspark.sql.session.SparkSession
-		"""
-		self.s3.set_spark(spark)
-
 	def load(self, spark=None):
 		return self.s3.load(path=self._path, spark=spark)
 
@@ -489,6 +482,9 @@ class S3Path:
 	def is_dir(self):
 		return self.s3.is_dir(path=self._path)
 
+	def make_directory(self):
+		return self.s3.mkdir(path=self._path)
+
 	@property
 	def size(self):
 		return self.s3.get_size(path=self._path)
@@ -499,5 +495,19 @@ class S3Path:
 	def delete(self):
 		return self.s3.delete(path=self._path)
 
+	def get_num_files(self):
+		if self.is_dir():
+			return len(self.s3.file_system.ls(path=self.path))
+		else:
+			raise NotADirectoryError(f'{self.path} is not a directory!')
+
+	def is_empty(self):
+		return self.get_num_files() == 0
+
 	dir = ls
 	list = ls
+	mkdir = make_directory
+	make_dir = make_directory
+	md = make_directory
+	is_directory = is_dir
+	rm = delete
